// Copyright 2015 Google Inc.

#import <GoogleCastRemoteDisplay/GCKVideoFrameInput.h>

#import <OpenGLES/gltypes.h>
#import <simd/simd.h>

@class EAGLContext;

GCK_ASSUME_NONNULL_BEGIN

/**
 * Specialization of GCKVideoFrameInput for OpenGL ES.
 *
 * This video frame input makes it easier to integrate an OpenGL ES renderer with Remote Display by
 * providing a fast (but somewhat inaccurate) color space conversion pipeline. Apps that wish to
 * have full control over this process must use the session object directly to dequeue video frame
 * buffers and enqueue them for encoding and transmission once rendered.
 *
 * Both OpenGL ES 2 and OpenGL ES 3 contexts are supported. OpenGL ES 1 contexts are not.
 *
 * @sa GCKRemoteDisplaySession
 *
 * @section gles_init Initialization
 *
 * Once a Remote Display session has been negotiated, initialize a frame input with it and set the
 * OpenGL ES context property. The frame input will synchronously create all of its required
 * resources in the property setter. Finally, set up the OpenGL ES rendering code for
 * Remote Display, creating a 2D texture matching the size of the Remote Display video stream as the
 * final render target. The texture format must be color renderable and must support bilinear
 * filtering. The texture must use the GL_TEXTURE_2D target.
 *
 * @section gles_produce Producing a frame
 *
 * To produce a Remote Display frame, first render the frame then call @ref encodeFrame: with the
 * render target texture. The frame input will encode 2 render passes to perform color space
 * conversion. Because OpenGL ES does not provide a callback mechanism for command buffer
 * completion, @ref syncAndSubmitFrame must eventually be called to synchronously wait for the frame
 * to finish processing.
 *
 * If the app has no further work to do after encoding a Remote Display frame, it may call
 * @ref syncAndSubmitFrame right away and block until the frame is ready. Otherwise, it is generally
 * safe to call @ref syncAndSubmitFrame at the beginning of the next frame. By that time, the
 * previously encoded frame will almost certainly be ready and the app will not block.
 *
 * @section gles_thread_safety Thread safety
 *
 * Because this frame input calls into OpenGL, it must be used on the same thread using its current
 * OpenGL ES context, or at least serially with any other thread using the context. To avoid race
 * conditions when objects of this class are deallocated, set the context property to nil on the
 * thread that owns the OpenGL ES context. This applies to every method and property get & set.
 *
 * @section gles_state_mutation State mutation
 *
 * When setting a context, encoding a frame or syncing and submitting a frame, OpenGL ES state must
 * be mutated. Whenever possible, the frame input saves and restores the global context state it
 * mutates as well as the thread's current context.
 *
 * @ingroup RemoteDisplay
 */
GCK_EXPORT
@interface GCKOpenGLESVideoFrameInput : GCKVideoFrameInput

/**
 * The OpenGL ES context to use for resource allocations. Changing this property reallocates
 * resources synchronously and can take some amount of time. Setting the context to nil will
 * immediately free all allocated resources.
 */
@property(nonatomic, GCK_NULLABLE) EAGLContext *context;

/**
 * Transform matrix applied to the texture coordinates of enqueued frames.
 *
 * This matrix is applied to the normalized texture coordinates used to sample enqueued frames
 * during the color space conversion rendering passes encoded by this frame input. It can be more
 * efficient to use this property than to perform the transform as a separate rendering pass.
 *
 * For example, to perform a y-flip, set the following matrix (here as an array of column vectors):
 * { {1.0, 0.0, 0.0}, {0.0, -1.0, 0.0}, {0.0, 1.0, 0.0} }
 *
 * Defaults to the identity matrix.
 */
@property(nonatomic) matrix_float3x3 textureCoordinateMatrix;

/**
 * Encode commands to process and transmit a frame.
 *
 * The frame input will encode 2 render passes to convert the frame to Y'CbCr with 4:2:0 chroma
 * subsampling. It will also encode a command buffer fence to test if the frame has finished
 * processing. This fence must eventually be tested and potentially waited on by calling
 * @ref syncAndSubmitFrame.
 *
 * The current OpenGL ES context is used and is assumed to be the frame input's context. If not,
 * behavior is undefined.
 *
 * If @ref syncAndSubmitFrame is not called before the next call to @ref encodeFrame:, the
 * previously encoded frame is dropped.
 *
 * If anything goes wrong, such as if the session runs out of video frame buffers or the texture
 * cannot be sampled, the frame is dropped.
 *
 * The texture's size must match the drawable size, must support bilinear filtering and must be a
 * 2D texture (using the GL_TEXTURE_2D target). The same texture can be used each frame since its
 * content does not need to remain unchanged.
 */
- (void)encodeFrame:(GLuint)texture;

/**
 * Wait for the previously encoded frame to be ready and submit it to the Remote Display session.
 *
 * Waits on the command buffer fence encoded by @ref encodeFrame: for up to 100 ms. If the fence is
 * signalled before this timeout, the frame is submitted to the session to be compressed and
 * transmitted. The presentation timestamp of the frame will be the result of CACurrentMediaTime
 * when the fence is signalled.
 *
 * If no frame is pending, returns immediately. If the fence is not signalled in time, the frame is
 * dropped.
 *
 * The current OpenGL ES context is used and is assumed to be the frame input's context. If not,
 * behavior is undefined.
 */
- (void)syncAndSubmitFrame;

@end

GCK_ASSUME_NONNULL_END
